﻿Imports System.Runtime.CompilerServices
Imports System.Reflection
Imports System.ComponentModel
Imports System.Data
Imports System.IO

Namespace System.Windows.Forms
   Public Module XDatasetAdapter
      <Extension()> _
      Public Function SuspendConstraints(dts As DataSet) As IDisposable
         Return dts.Attach.SuspendConstraints.Enter
      End Function
      <Extension()> _
      Public Sub RegisterSaveCallback(dts As DataSet, dlg As Action, Optional value As Boolean = True)
         dts.Attach.RegisterSaveCallback(dlg, value)
      End Sub
      <Extension()> _
      Public Function Attach(dts As DataSet) As DataTableAttach.DatasetAttach
         Return dts.Tables(0).Attach.Dataset_Attach
      End Function
      <Extension()> _
      Public Function Adapter(dts As DataSet) As DatasetAdapterBase
         Adapter = Nothing
         If DatasetAdapterBase.Attachs.TryGetValue(dts, Adapter) Then Exit Function
         Adapter = New DatasetXmlAdapter(dts)
         DatasetAdapterBase.Attachs.Add(dts, Adapter)
      End Function
      <Extension()> _
      Public Function DataFile(dts As DataSet, path As String) As DataSet
         DirectCast(dts.Adapter, DatasetXmlAdapter).DataFile = New FileInfo(path)
         Return dts
      End Function
      <Extension()> _
      Public Function DataFile(dts As DataSet) As FileInfo
         Dim att = DirectCast(dts.Adapter, DatasetXmlAdapter)
         If att.Null Then Return Nothing
         Return att.DataFile
      End Function

      ''' <summary>Clear dataset and refill. Optional only the given tables (not supported in dataset-only-scenarios)</summary>
      <Extension()> _
      Public Sub Fill(dts As DataSet, ParamArray tables() As DataTable)
         dts.Adapter.Fill(tables)
      End Sub
      <Extension()> _
      Public Sub Save(dts As DataSet, frm As Form, Optional checkForChanges As Boolean = True)
         dts.Adapter.Save(frm, checkForChanges)
      End Sub

      <Extension()> _
      Public Function HandleFormClosing(dts As DataSet) As FormClosingEventHandler
         Return dts.Adapter.HandleFormClosing
      End Function

#Region "Winfom-Close-Bug"         ' Aufwändig gestaltete DGVs schmeißen manchmal Fehler beim Disposen des Forms.

      Private Sub Form_Closed(s As Object, e As EventArgs)
         'manchmal schmeissen DGVs Fehler beim Schließen des Forms. Ist schlecht reproduzierbar.
         'Reproduziere: Starte RaslDb, gehe auf Artikel-Tab und Form schließen
         With DirectCast(s, Form)
            If .DialogResult <> DialogResult.None Then Return
            .Validate() 'nochn Bug: DGV kriegt beim Disposen einen ZeilenIndex-Fehler, wenns sich die ZufügeZeile im Edit-Modus befindet
            .SuspendLayout()
#If True Then            'geringster Eingriff
            For Each ctl In New GetChilds(Of Control)(Function(c) c.Controls).AllAsList(.Self)
               With TryCast(ctl, DataGridView)
                  If .NotNull Then .Parent.SuspendLayout() : .Dispose()
               End With
            Next
#ElseIf True Then            'einfachster Eingriff
            With .Controls
               'Controls könnten sich theor. auch gegenseitig disposen, über Events. dieses hier disposed ganz sicher immer das letzte Element was noch da ist
               While .Count > 0 : .Item(.Count - 1).Dispose() : End While
            End With
#ElseIf True Then            'kein Eingriff
         Return
#ElseIf True Then
#End If
         End With
      End Sub

      'Beim Schließen des MainForms wird Form_Closed der anneren Forms garnet mehr aufgerufen. Daher hier alle alle Controls disposen
      Private Sub MainForm_Closed(s As Object, e As EventArgs)
         For Each frm As Form In Application.OpenForms
            Form_Closed(frm, e)
         Next
      End Sub

#End Region 'Winfom-Close-Bug

      Friend _KnownDatasets As New Dictionary(Of Type, DataSet)

      Private _FormClosedKey As Object = GetType(Form) _
         .GetField("EVENT_FORMCLOSED", BindingFlags.NonPublic Or BindingFlags.Static) _
         .GetValue(Nothing)

      Private Sub ThrowIfFormIsAlreadyRegistered(frm As Form)
         ' mit Reflection den Form_Closed-Delegaten auspuhlen, und gucken, ob 'mein' Form_Closed - Handler bereits eingetragen ist. Falls ja - Exception
         Dim events = DirectCast(DataTableAttach.EventsFieldInfo.GetValue(frm), EventHandlerList)
         Dim formClosed = events(_FormClosedKey)
         With events(_FormClosedKey)
            If .NotNull AndAlso .GetInvocationList.Contains(DirectCast(AddressOf Form_Closed, FormClosedEventHandler)) Then
               Throw New InvalidOperationException("can't register the same Form twice!")
            End If
         End With
      End Sub

      ''' <summary>
      ''' registers all datasets of this type and all bindingsources of the form, to improve performance and to avoid data-redundance
      ''' </summary>
      <Extension()> _
      Public Function Register(Of T As DataSet)(dts As T, frm As Form, handleFormClosing As Boolean) As T
         ThrowIfFormIsAlreadyRegistered(frm)
         AddHandler frm.FormClosed, AddressOf Form_Closed 'EVENT_FORMCLOSED
         If handleFormClosing Then AddHandler frm.FormClosing, dts.Adapter.HandleFormClosing
         Return Register(frm, dts)
      End Function

      ''' <summary>
      ''' registers all datasets of this type and all bindingsources of the ContainerControl, to improve performance and to avoid data-redundance
      ''' </summary>
      <Extension()> _
      Public Function Register(Of T As DataSet)(dts As T, ctlContainer As ContainerControl) As T
         Dim frm = TryCast(ctlContainer, Form)
         If frm.NotNull Then Return dts.Register(frm, False)
         Return Register(ctlContainer, dts)
      End Function

#If True Then


      Private Function Register(Of T As DataSet)(ctlRoot As ContainerControl, dts As T) As T
#If False Then
durchsucht rekursiv alle userdefined Felder des Forms/UserControl
1a) ersetzt alle T-Instanzen durch das einzig wahre dts
1b) disposet die üflüssigen T-Instanzen 
2a) registriert alle T-Bindingsources bei den TableAttaches ihrer DataTable
2b) stöpselt alle T-Bindingsources auf das einzig wahre um. Inkonsistente Zwischenzustände werden überbrückt durch SetIniting(True/False) 
3) stöpselt alle Controls, die eine Datasource haben mit einem Wert des Typs T auf das einzig wahre um
#End If
         Dim tpDts = GetType(T)
         Dim tmp As DataSet = Nothing
         If _KnownDatasets.TryGetValue(tpDts, tmp) Then dts = DirectCast(tmp, T) Else _KnownDatasets.Add(tpDts, dts)
         Register = dts
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public Or BindingFlags.DeclaredOnly _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         Dim bss = New List(Of BindingSource)
         Dim dataSourceMatch As Func(Of Object, Boolean) = Function(scr) TypeOf scr Is T OrElse scr Is GetType(T)
         Dim oldDatasets As New HashSet(Of T)({dts})
         Dim recurse As Action(Of Control) =
            Sub(ctl)
               ctl.Controls.Cast(Of Control).ForEach(recurse)
               If Not (TypeOf ctl Is Form OrElse TypeOf ctl Is UserControl) Then Return
               For Each fld In ctl.GetType.GetFields(bindFlags)
                  If fld.FieldType Is tpDts Then                                                               ' 1)
                     If oldDatasets.Add(DirectCast(fld.GetValue(ctl), T)) Then fld.SetValue(ctl, dts) ' 1a)
                  ElseIf fld.FieldType Is GetType(BindingSource) Then                                          ' 2)
							Dim bs = DirectCast(fld.GetValue(ctl), BindingSource)
							If bs.Null Then Continue For
							Dim finalDataSource = bs.DataSource
                     Do
                        With TryCast(finalDataSource, BindingSource)
                           If .Null Then Exit Do
                           finalDataSource = .DataSource
                        End With
                     Loop
                     If Not dataSourceMatch(finalDataSource) Then Continue For
                     bss.Add(bs)
                     If bs.DataMember = "" Then Continue For
                     Dim tb = If(dts.Tables(bs.DataMember), dts.Relations(bs.DataMember).ChildTable)
                     tb.Attach.AddSource(bs)                                                                ' 2a)
                  ElseIf GetType(Control).IsAssignableFrom(fld.FieldType) Then                              ' 3)
                     Dim prp = fld.FieldType.GetProperty("DataSource")
                     If prp.Null Then Continue For
                     Dim itm = fld.GetValue(ctl)
                     If itm IsNot Nothing AndAlso dataSourceMatch(prp.GetValue(itm, Nothing)) Then
                        If prp.GetValue(itm, Nothing) IsNot dts Then prp.SetValue(itm, dts, Nothing)
                     End If
                  End If
               Next
            End Sub
         recurse(ctlRoot)
         oldDatasets.Remove(dts)
         For Each ds In oldDatasets : ds.Dispose() : Next                                                   ' 1b)
         For Each bs In bss                                                                                 ' 2b)
            bs.SetIniting(True)
            If dataSourceMatch(bs.DataSource) AndAlso bs.DataSource IsNot dts Then bs.DataSource = dts
         Next
         For Each bs In bss : bs.SetIniting(False) : Next
      End Function

#ElseIf True Then
      Private Function Register(Of T As DataSet)(frm As ContainerControl, dts As T) As T
#If False Then
durchsucht alle Felder des Forms
1a) ersetzt alle T-Instanzen durch das einzig wahre dts
1b) disposet die üflüssigen T-Instanzen 
2a) registriert alle T-Bindingsources bei den TableAttaches ihrer DataTable
2b) stöpselt alle T-Bindingsources auf das einzig wahre um. Inkonsistente Zwischenzustände werden überbrückt durch SetIniting(True/False) 
3) stöpselt alle Controls, die eine Datasource haben mit einem Wert des Typs T auf das einzig wahre um
#End If
         Dim tpDts = GetType(T)
         Dim tmp As DataSet = Nothing
         If _KnownDatasets.TryGetValue(tpDts, tmp) Then dts = DirectCast(tmp, T) Else _KnownDatasets.Add(tpDts, dts)
         Register = dts
         Dim allCtls = New GetChilds(Of Control)(Function(ctl) ctl.Controls).AllAsList(frm)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         Dim bss = New List(Of BindingSource)
         Dim dataSourceMatch As Func(Of Object, Boolean) = Function(scr) TypeOf scr Is T OrElse scr Is GetType(T)
         Dim oldDatasets As New HashSet(Of T)({dts})

         For Each ctl In allCtls.OfType(Of ContainerControl)()
            For Each fld In ctl.GetType.GetFields(bindFlags)
               If fld.FieldType Is tpDts Then                                                               ' 1)
                  If oldDatasets.Add(DirectCast(fld.GetValue(ctl), T)) Then fld.SetValue(ctl, dts) ' 1a)
               ElseIf fld.FieldType Is GetType(BindingSource) Then                                          ' 2)
                  Dim bs = DirectCast(fld.GetValue(ctl), BindingSource)
                  Dim finalDataSource = bs.DataSource
                  Do
                     With TryCast(finalDataSource, BindingSource)
                        If .Null Then Exit Do
                        finalDataSource = .DataSource
                     End With
                  Loop
                  If Not dataSourceMatch(finalDataSource) Then Continue For
                  bss.Add(bs)
                  If bs.DataMember = "" Then Continue For
                  Dim tb = If(dts.Tables(bs.DataMember), dts.Relations(bs.DataMember).ChildTable)
                  tb.Attach.AddSource(bs)                                                                ' 2a)
               ElseIf GetType(Control).IsAssignableFrom(fld.FieldType) Then                              ' 3)
                  Dim prp = fld.FieldType.GetProperty("DataSource")
                  If prp.Null Then Continue For
                  Dim itm = fld.GetValue(ctl)
                  If itm IsNot Nothing AndAlso dataSourceMatch(prp.GetValue(itm, Nothing)) Then prp.SetValue(itm, dts, Nothing)
               End If
            Next
         Next
         oldDatasets.Remove(dts)
         For Each ds In oldDatasets : ds.Dispose() : Next                                                   ' 1b)
         For Each bs In bss                                                                                 ' 2b)
            bs.SetIniting(True)
            If dataSourceMatch(bs.DataSource) AndAlso bs.DataSource IsNot dts Then bs.DataSource = dts
         Next
         For Each bs In bss : bs.SetIniting(False) : Next
      End Function

#ElseIf True Then
      Private Function Register(Of T As DataSet)(frm As ContainerControl, dts As T) As T
         Dim tpDts = dts.GetType
         Dim tmp As DataSet = Nothing
         Dim wasKnown = _KnownDatasets.TryGetValue(tpDts, tmp)
         If wasKnown Then dts = DirectCast(tmp, T)
         Register = dts
         Dim allCtls = New GetChilds(Of Control)(Function(ctl) ctl.Controls).AllAsList(frm)
         Dim dgvSuspend = New GridSuspend(dts, allCtls)
         Dim oldDatasets As New HashSet(Of DataSet)
         Dim shouldReplace As Func(Of Object, Boolean) = Function(src) _
            src IsNot Nothing AndAlso _
            (src Is tpDts OrElse src.GetType Is tpDts AndAlso src IsNot dts)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         For Each ctl In allCtls.OfType(Of ContainerControl)()
            Dim fields = ctl.GetType.GetFields(bindFlags)
            For Each fld In fields
               If fld.FieldType Is tpDts Then
                  oldDatasets.Add(DirectCast(fld.GetValue(ctl), DataSet))
                  fld.SetValue(ctl, dts)
                  Continue For
               End If
               If Not GetType(IComponent).IsAssignableFrom(fld.FieldType) Then Continue For
               Dim prp = fld.FieldType.GetProperty("DataSource")
               If prp.Null Then Continue For
               Dim itm = fld.GetValue(ctl)
               If itm Is Nothing Then Continue For
               TryRegisterAsBindingSource(itm, dts, tpDts)
               If shouldReplace(prp.GetValue(itm, Nothing)) Then prp.SetValue(itm, dts, Nothing)
            Next
         Next
         dgvSuspend.Resume()
         oldDatasets.Remove(dts)
         For Each ds In oldDatasets : ds.Dispose() : Next
         If Not wasKnown Then _KnownDatasets.Add(tpDts, dts)
      End Function

      Private Sub TryRegisterAsBindingSource(itm As Object, dts As DataSet, tpDts As Type)
         Dim bs = TryCast(itm, BindingSource)
         If bs.Null Then Return
         Dim dataMember = bs.DataMember
         Dim dataSource = bs.DataSource
         If dataSource Is Nothing Then Return
         Do
            With TryCast(dataSource, BindingSource)
               If .Null Then Exit Do
               If dataMember = "" Then dataMember = .DataMember
               dataSource = .DataSource
            End With
         Loop
         If dataSource IsNot tpDts AndAlso dataSource.GetType IsNot tpDts Then Return
         With dts.Tables(dataMember)
            If .NotNull Then .Attach.AddSource(bs) : Return
         End With
         With dts.Relations(dataMember)
            If .NotNull Then .ChildTable.Attach.AddSource(bs)
         End With
      End Sub

      Private Class GridSuspend
         'beim Umstöpseln der BindingSources die DGVs abhängen
         Private _Tuples As New List(Of Tuple(Of DataGridView, BindingSource))

         Public Sub New(dts As DataSet, allCtls As List(Of Control))
            If dts.Tables.Cast(Of DataTable).All(Function(tb) tb.Rows.Count = 0) Then Return
            For Each dgv In allCtls.OfType(Of DataGridView)()
               Dim bs = TryCast(dgv.DataSource, BindingSource)
               If bs.Null Then Continue For
               _Tuples.Add(Tuple.Create(dgv, bs))
               dgv.DataSource = Nothing
            Next
         End Sub
         Public Sub [Resume]()
            For Each tpl In _Tuples
               tpl.Item1.DataSource = tpl.Item2
            Next
         End Sub
      End Class
#ElseIf True Then

#End If


   End Module
End Namespace

