﻿#Region "FileHeader"
#If False Then
Zustandsbehaftete "Behaviors" für DatagridView (zustands**freie** Behaviors siehe in "DatagridViewBehavior.vb")

• DrawGroupLines(): In sortierten Spalten stehen gleiche Werte zusammen - kann man als Gruppe auffassen. DrawGroupLines() malt eine Trennlinie über der obersten Zelle jeder Gruppe. Zusätzlich sind standardmäßig die Wiederholungen ausgeblendet, sodass die Gruppierung gut erkennbar wird. Besonders reizvoll bei hierarchischer Sortierung nach 2 Spalten - da treten Gruppen und Untergruppen zutage.
  Bei normaler Sortierung werden die Trennlinien nicht nur in die Sortierspalte eingezeichnet, sondern überspannen alle Spalten

• HierarchicalSort(): bewirkt, dass 2-spaltig hierarchisch sortiert wird, wenn man nacheinander mehrere ColumnHeaders klickst. Dabei wird die letzt-geklickste Spalte jeweils zur Haupt-Sortier-Spalte 

• RestoreCurrent(): Beim Umschalten der Sortierung wird der vorher selektete Datensatzes an seiner neuen Position wieder aufgesucht

• UnFocusedSelectionColor(): SelectionBackColor wird auf Khaki geändert. Und diese  wird bei Focus-Verlust etwas aufgehellt, sodass man bei mehreren DGVs immer noch erkennen kann, welches den Focus hat.

#### HierarchicalSort() und RestoreCurrent() sind bischen Fakes, denn sie reichen den Aufruf nur durch an die BindingSource.Attach-Klasse ####

#End If '-- Options, Imports
Imports System.ComponentModel
Imports System.Runtime.CompilerServices
#End Region 'FileHeader


Namespace System.Windows.Forms

   Public Class DataGridViewAttach

      Friend Shared ReadOnly Attaches As New ComponentProperty(Of DataGridView, DataGridViewAttach)(Function(dgv) New DataGridViewAttach(dgv))

      Private WithEvents _Dgv, _dgvGrouping, _dgvHideRepeat, _UnFocusedSelectionColor As DataGridView
      Public Sub New(dgv As DataGridView)
         Me._Dgv = dgv
      End Sub

		Public Function DrawGroupLines(groupCols As DataGridViewColumn(), Optional drawLines As Boolean = True, Optional hideRepeat As Boolean = True) As DataGridViewAttach
			_CellComparer = AddressOf CellComparerDefault
			_dgvGrouping = If(drawLines, _Dgv, Nothing)
			If drawLines Then
				_Dgv.CellBorderStyle = DataGridViewCellBorderStyle.None
				_SortedCols = groupCols
			End If
			_dgvHideRepeat = If(drawLines AndAlso hideRepeat, _Dgv, Nothing)
			Return Me
		End Function
		Public Function DrawGroupLines(groupCols As DataColumn(), cellEqualComparer As Func(Of DataGridViewCell, DataGridViewCell, Boolean), Optional drawLines As Boolean = True, Optional hideRepeat As Boolean = True) As DataGridViewAttach
         _CellComparer = cellEqualComparer
         _dgvGrouping = If(drawLines, _Dgv, Nothing)
         If drawLines Then
            _Dgv.CellBorderStyle = DataGridViewCellBorderStyle.None
            If groupCols.NotNull Then : SetSortedCols(groupCols.Select(Function(col) col.ColumnName))
            Else : SetSortedCols(DirectCast(_Dgv.DataSource, BindingSource))
            End If
         End If
         _dgvHideRepeat = If(drawLines AndAlso hideRepeat, _Dgv, Nothing)
         Return Me
      End Function

      Public Function DrawGroupLines(groupCols As DataColumn(), Optional drawLines As Boolean = True, Optional hideRepeat As Boolean = True) As DataGridViewAttach
         Return DrawGroupLines(groupCols, AddressOf CellComparerDefault, drawLines, hideRepeat)
      End Function
      Public Function DrawGroupLines(Optional drawLines As Boolean = True, Optional hideRepeat As Boolean = True) As DataGridViewAttach
			Return DrawGroupLines(Nothing, AddressOf CellComparerDefault, drawLines, hideRepeat)
		End Function
      Public Function HierarchicalSort(Optional value As Boolean = True) As DataGridViewAttach
         DirectCast(_Dgv.DataSource, BindingSource).Attach.HierarchicalSort(value)
         Return Me
      End Function
      Public Function RestoreCurrent(Optional value As Boolean = True) As DataGridViewAttach
         DirectCast(_Dgv.DataSource, BindingSource).Attach.RestoreCurrent(value)
         Return Me
      End Function
      Public Function UnFocusedSelectionColor(Optional value As Boolean = True) As DataGridViewAttach
         _UnFocusedSelectionColor = If(value, _Dgv, Nothing)
         If value Then
            _Dgv.DefaultCellStyle.SelectionForeColor = _Dgv.DefaultCellStyle.ForeColor
            _UnFocusedSelectionColor_FocusChanged(Nothing, Nothing)
         End If
         Return Me
      End Function

      Private _SortedCols As DataGridViewColumn() = {}
      Private Shared _QuoteChars As Char() = "[]".ToCharArray

      Private Sub SetSortedCols(names As IEnumerable(Of String))
         Dim dgvCols = _Dgv.Columns.Cast(Of DataGridViewColumn).ToArray(Function(col) col.DataPropertyName)
         Dim nms = names.ToArray
         _SortedCols = (From name In names Distinct Join col In _Dgv.Columns.Cast(Of DataGridViewColumn)() On col.DataPropertyName Equals name Select col).ToArray
      End Sub
      Private Sub SetSortedCols(bindingSource As BindingSource)
         If bindingSource.WaitForGetReady(AddressOf SetSortedCols) Then Return
         Dim dv = DirectCast(bindingSource.List, DataView)
         If dv.Sort = "" Then _SortedCols = {} : Return
         SetSortedCols(dv.Sort.Split(","c).Select(Function(s) s.Split(_QuoteChars)(1)))
      End Sub

      Private Shared Function CellComparerDefault(cell1 As DataGridViewCell, cell2 As DataGridViewCell) As Boolean
         Return Object.Equals(cell1.Value, cell2.Value)
      End Function
      Private _CellComparer As Func(Of DataGridViewCell, DataGridViewCell, Boolean)

      Private Function IsGroupHeader( i As Integer,  rowIndex As Integer) As Boolean
         If rowIndex >= _Dgv.RowCount Then Return False
         Return rowIndex <= 0 OrElse Not _CellComparer(_Dgv(i, rowIndex - 1), _Dgv(i, rowIndex))
      End Function

      Private Sub _HideRepeat_CellFormatting(sender As Object, e As DataGridViewCellFormattingEventArgs) Handles _dgvHideRepeat.CellFormatting
         If e.RowIndex < 1 Then Return
         If Not _SortedCols.Contains(_Dgv.Columns(e.ColumnIndex)) Then Return
         If IsGroupHeader(e.ColumnIndex, e.RowIndex) Then Return
         e.Value = "" : e.FormattingApplied = True
      End Sub

      Private Sub Grouping_RowPostPaint(sender As Object, e As DataGridViewRowPostPaintEventArgs) Handles _dgvGrouping.RowPostPaint
         Dim rowIndex = e.RowIndex
         If rowIndex < 1 Then Return
         Select Case _SortedCols.Length
            Case 0 : Return
            Case 1 : If IsGroupHeader(_SortedCols(0).Index, rowIndex) Then DrawTop(_dgvGrouping, e.Graphics, rowIndex)
            Case Else
               For Each col In _SortedCols
                  If IsGroupHeader(col.Index, rowIndex) Then DrawTop(_dgvGrouping, e.Graphics, rowIndex, col.Index)
               Next
         End Select
      End Sub

      Private Sub DrawTop(dgv As DataGridView, g As Graphics, rowIndex As Integer, Optional columnIndex As Integer? = Nothing)
         Dim rct = dgv.GetCellDisplayRectangle(If(columnIndex, dgv.DisplayedColumnCount(True) - 1), rowIndex, True)
         With rct
            If .IsEmpty Then Return
            If Not columnIndex.HasValue Then
               Dim x = If(dgv.RowHeadersVisible, dgv.RowHeadersWidth, 0)
               .Width = .Right - x
               .X = x
            End If
            g.DrawLine(Pens.Red, .X, .Y, .Right, .Y)
         End With
      End Sub

      Private Sub Grouping_Sorted(sender As Object, e As EventArgs) Handles _dgvGrouping.Sorted
         SetSortedCols(DirectCast(_Dgv.DataSource, BindingSource))
      End Sub

      Private Sub _UnFocusedSelectionColor_FocusChanged( sender As Object,  e As EventArgs) Handles _UnFocusedSelectionColor.GotFocus, _UnFocusedSelectionColor.LostFocus
         _Dgv.DefaultCellStyle.SelectionBackColor = If(_Dgv.Focused, Color.Khaki, Color.FromArgb(240, 240, 180))
      End Sub

		Public Function GetSelectedRows(Of T As DataRow)() As IEnumerable(Of T)
			Dim selRows = From cll In _Dgv.SelectedCells.Cast(Of DataGridViewCell) Where cll.RowIndex >= 0 Select cll.OwningRow Distinct
			Return selRows.Select(Function(rw) DirectCast(DirectCast(rw.DataBoundItem, DataRowView).Row, T))
		End Function

	End Class

   <EditorBrowsable(EditorBrowsableState.Never)> _
   Public Module DataGridViewX
      <Extension> _
      Public Function Attach(dgv As DataGridView) As DataGridViewAttach
         Return DataGridViewAttach.Attaches(dgv)
      End Function
   End Module


End Namespace